C
      PROGRAM DGS16C
C
C     Digisonde 256 16 Channel data decoder
C
C     This program forms the basis of a decoder for 16 channel (16C) 
C     data obtained from a UMass Lowell Digisonde 256 ionosonde.
C
C     Terence Bullett
C     Air Force Research Laboratory
C     15 September 2001
C     linux g77
C
C     License:
C     This is in the public domain.  
C     This software was paid for by the Taxpayers of the
C     United States of America.
C     Thank one the next time you meet.
C
C     References:
C        GL-TR-09-0190 "ARTIST Tape Output Formats"  July 1990  J. Tang, et al
C
C        "Decoding of Status Information in MMM Digisonde Ionograms"
C                 I. Galkin  September 2001
C
C        Digisonde 256 -- General Description of the Compact Digital
C             Ionospheric Sounder", Bibl, Reinisch, Kitrosser, 1981 Revised
C              aka "Green Book"
C
C        "Digisonde 256 Operators Manual"  Terence Bullett  
C
C     Useage:
C        dgs16c inputfilename {options}
C        decoded data are written in ASCII to stdout
C
C        The input file must exist and contain 16 channel D256 data.
C
C     Revision History
C     17Jan02 TWB - Added status interpetation code.
C     12Feb02 TWB - Added output formats for gnuplot
C     03Mar02 TWB - Added simple threshold filtering
C     09Mar02 TWB - Corrected Doppler Table, improved implementation
C                 - Version 0.10 ready for testing
C
C     Bugs
C
C     Limitations
C        I can't reliably read binary files from stdin, at least not in any way
C        that is remotely portable.  Thus this program cannot be a pipe.
C        Sorry.
C
C     To Do
C      - *** 256 range gate modes ***
C      - *** Precision Group Height modes ****
C      - Improve the use of custom antenna tables
C      - Allow for a range of frequencies to be extracted
C      - Test and Debug !!!!
C
C========================================================================
C
C      Variables
      CHARACTER*120 CLP,INFILE
      INTEGER I,K,NA,IU,NBEAM,NDOPP
      INTEGER*1 IBUF(4096)
      LOGICAL VERBOSE,INFOK,EOF,FRQ_SEARCH,OK
      LOGICAL WRAMP,WRPHA,WRSPL,FIXPHAS,THRESH 
      REAL EPSILON,DELFRQ,WANTEDFRQ,VERSION,AMPTHR
C
      REAL FREQ, EAMPS(256,16), EPHASE(256,16), RGT(256)
      REAL FNOISE
      CHARACTER*17 TIME
      INTEGER*1 PREFACE(57)
      INTEGER MAXRBIN
C
      REAL APOL(16),HDOP(16),AZ(16),ZN(16)
      INTEGER ICS(16)
C
C     Functions
      LOGICAL READ4KB, UNPACK_16C
      REAL AMPV
C
      PARAMETER (EPSILON=1.0E-6, DELFRQ=0.02, VERSION=0.10)

C
C      Parse command line options
C
C
C     Open input file
C     NOTE: This is operating system and compiler dependent.
C           You will likely have to edit this.
C     Pre-opened unit numbers under g77:  stdin=5, stdout=6, stderr=0
C        
C     Read the command line for arguments
      K = IARGC() 
C     Filename must exist
      IF (K.LT.1) THEN
         WRITE(0,*) "dgs16c-> ERROR: Input file must be specified"
         CALL EXIT(-1)
         RETURN
      ENDIF
C
C     Initialize variables
      INFILE='error'
      IU=5
      VERBOSE=.FALSE.
      FRQ_SEARCH=.FALSE.
      WANTEDFREQ=0.0
      WRAMP=.FALSE.
      WRPHA=.FALSE.
      WRSPL=.FALSE.
      FIXPHAS=.FALSE.
      THRESH=.FALSE.

C     Parse the command line.  Not very robust here, but it works. Mostly.
      DO NA = 1, K
        CALL GETARG(NA,CLP)
        IF (CLP.EQ.'-v') THEN
          VERBOSE = .TRUE.
        ELSE IF (CLP.EQ.'-h') THEN
          WRITE(*,*)' useage: dgs16c {-vapsh} {-f #} filename'
          WRITE(*,*)' '
          WRITE(*,*)'	Decode 16 channel data from a Digisonde 256 16C file'
          WRITE(*,*)' Filename must be provided on the command line.'
          WRITE(*,*)' Command line options:'
          WRITE(*,*)'  -v  Verbose.  Prints the status on stdout'
          WRITE(*,*)'  -f  <freq>  extract data for <frequency>'
          WRITE(*,*)'  -c  Correct phase based on range gate'
          WRITE(*,*)'  -a  Write Amplitude Table'
          WRITE(*,*)'  -p  Write Phase Table'
          WRITE(*,*)'  -s  Write data for gnuplot splot command'
          WRITE(*,*)'  -t  <val> Threshold Detect, <val> above noise'
          WRITE(*,*)'      output format set to -s'
          WRITE(*,*)' '      
          WRITE(*,*)' Returns exit status as follows:'
          WRITE(*,*)'  0:  No Problems'
          WRITE(*,*)' -1:  File unreadable or nonexistent'
          RETURN 
        ELSE IF (CLP.EQ.'-f') THEN
           FRQ_SEARCH = .TRUE.
           CALL GETARG(NA+1,CLP)
           READ(CLP,*) WANTEDFRQ
        ELSE IF (CLP.EQ.'-a') THEN
           WRAMP = .TRUE.
        ELSE IF (CLP.EQ.'-p') THEN
           WRPHA=.TRUE.
        ELSE IF (CLP.EQ.'-s') THEN
           WRSPL=.TRUE.    
        ELSE IF (CLP.EQ.'-c') THEN
           FIXPHAS=.TRUE.
        ELSE IF (CLP.EQ.'-t') THEN
           WRSPL = .FALSE.
           THRESH = .TRUE.
           CALL GETARG(NA+1,CLP)
           READ(CLP,*) AMPTHR
        ELSE    
          INFILE=CLP
        ENDIF
      ENDDO      
C
C     Initialize the ZTL Rx antenna tables.
      CALL ZTL_INIT()
C
C     Check if input file exists
      INQUIRE(FILE=INFILE,EXIST=INFOK)
      IF (.NOT.INFOK) THEN
         CALL EXIT(-1)
      ENDIF
C
C     Open input file.  This is for g77.  You mileage will vary.
      OPEN(UNIT=IU,FILE=INFILE, ACCESS='DIRECT', FORM='UNFORMATTED', 
     +     STATUS='OLD', RECL=4096 )
C
      EOF=READ4KB(IU,IBUF)
      DO WHILE (.NOT.EOF)
        OK=UNPACK_16C(IBUF,  
     +               TIME,PREFACE,FREQ,EAMPS,EPHASE,RGT,MAXRBIN)
        IF (.NOT.OK) CYCLE
C
        IF(FRQ_SEARCH.AND.(ABS(FREQ-WANTEDFRQ).GT.DELFRQ)) THEN
C          Read next record.  This will have to be more sophisticated for
C          Recordtype=  (2 frequencies per record, amplitude only)
           EOF=READ4KB(IU,IBUF)
           CYCLE
        ENDIF
C
C       Optionally fix the phases.
        IF (FIXPHAS) CALL FIXPHASE(RGT,EPHASE,MAXRBIN)
C
C       Determine the channel meta data from the preface
        CALL CHANNEL_MAP (PREFACE, APOL,HDOP,AZ,EL,NBEAM,NDOPP)
C
C       Sort the channels to change the output order
        CALL SORT_SPECTRA(APOL,HDOP,AZ,EL,4,  ICS)
C
C       Calculate the 'noise floor'
        FNOISE = AMPV(EAMPS,MAXRBIN)
C
        IF (WRSPL) THEN
C          Write the data in an gnuplot splot compatable format
C          For each data point:
           WRITE(*,110) 'Time','Freq','Range','Polarize','Noise',
     +                   'Doppler','Az','Zn','Ampltude','Phase'
           DO I=1, MAXRBIN
            DO J=1,16
              WRITE(*,111) TIME,FREQ,RGT(I),APOL(ICS(J)),FNOISE,
     +                     HDOP(ICS(J)),AZ(ICS(J)),ZN(ICS(J)),
     +                     EAMPS(I,ICS(J)),EPHASE(I,ICS(J))
            ENDDO
            WRITE(*,*) ''
           ENDDO
           WRITE(*,*) ''
        ELSE IF (THRESH) THEN
C          Write the data in an gnuplot splot compatable format
C          For each data point that passes the AMPTHR threshold:
           WRITE(*,110) 'Time','Freq','Range','Polarize','Noise',
     +                   'Doppler','Az','Zn','Ampltude','Phase'
           DO I=1, MAXRBIN
            DO J=1,16
             IF (EAMPS(I,ICS(J)).GT.(FNOISE+AMPTHR)) THEN 
                WRITE(*,111) TIME,FREQ,RGT(I),APOL(ICS(J)),FNOISE,
     +                       HDOP(ICS(J)),AZ(ICS(J)),ZN(ICS(J)),
     +                       EAMPS(I,ICS(J)),EPHASE(I,ICS(J))
             ENDIF
            ENDDO
            WRITE(*,*) ''
           ENDDO
           WRITE(*,*) ''

        ELSE IF (WRAMP.AND.WRPHA) THEN
C          Write the amplitude and phase tables
           WRITE(*,106) TIME, FREQ, (PREFACE(I),I=1,57)
           WRITE(*,104) 'Polarize',(APOL(ICS(I)),I=1,16)
           WRITE(*,105) 'Doppler',(HDOP(ICS(I)),I=1,16)
           WRITE(*,104) 'Azimuth',(AZ(ICS(I)),I=1,16)
           WRITE(*,104) 'Zenith',(ZN(ICS(I)),I=1,16)
           DO I=1,MAXRBIN
            WRITE(*,103) FREQ,RGT(I),(EAMPS(I,ICS(J)),J=1,16),
     +                              (EPHASE(I,ICS(J)),J=1,16)
           ENDDO
           WRITE(*,*) ''

C          Write both amplitude and phase tables
        ELSE IF (WRAMP) THEN
C          Write just the amplitude table
           WRITE(*,106) TIME, FREQ, (PREFACE(I),I=1,57)
           WRITE(*,104) 'Polarize',(APOL(ICS(I)),I=1,16)
           WRITE(*,105) 'Doppler',(HDOP(ICS(I)),I=1,16)
           WRITE(*,104) 'Azimuth',(AZ(ICS(I)),I=1,16)
           WRITE(*,104) 'Zenith',(ZN(ICS(I)),I=1,16)
           DO I=1,MAXRBIN
            WRITE(*,103) FREQ,RGT(I),(EAMPS(I,ICS(J)),J=1,16)
           ENDDO
           WRITE(*,*) ''
C
        ELSE IF (WRPHA) THEN
C          Write just the phase table
           WRITE(*,106) TIME, FREQ, (PREFACE(I),I=1,57)
           WRITE(*,104) 'Polarize',(APOL(ICS(I)),I=1,16)
           WRITE(*,105) 'Doppler',(HDOP(ICS(I)),I=1,16)
           WRITE(*,104) 'Azimuth',(AZ(ICS(I)),I=1,16)
           WRITE(*,104) 'Zenith',(ZN(ICS(I)),I=1,16)
           DO I=1,MAXRBIN
            WRITE(*,103) FREQ,RGT(I),(EPHASE(I,ICS(J)),J=1,16)
           ENDDO
           WRITE(*,*) ''
        ENDIF
C
C       Read the next block of data
        EOF=READ4KB(IU,IBUF)
      ENDDO
         
 100  FORMAT (25(1X,Z2.2))
 101  FORMAT (100(1X,Z1.1))
 102  FORMAT (F5.1,3X,16F6.1, ' # ', 16F6.1)
 103  FORMAT (F8.4,F8.1,32F8.1)
 104  FORMAT ('#',A15,32F8.2)
 105  FORMAT ('#',A15,32F8.3)
 106  FORMAT ('# ',A17,F7.4,1X,11Z1,1X,8Z1,1X,6Z1,1X,7Z1,1X,5Z1,1X,3Z1,
     +        1X,3Z1,1X,4Z1,1X,3Z1,1X,3Z1,1X,4Z1)
 110  FORMAT ('#',A16,10A10)
 111  FORMAT (A17,F10.4,10F10.3)
      END

C
C====+==================================================================+==
C
      REAL FUNCTION AMPV(EAMPS,MAXRBIN)
C
C     Calculate the Amplitude Most Pobable Value as a proxy for the
C     noise floor of the Range-Doppler data in EAMPS
C     Data are expected in the range of 0 to 120 dB and binned in 2dB bins.
C     The resolution, range and number of bins is hard-coded.  Sorry.
C
      REAL EAMPS(256,16),XMIN,XMAX,DX
      INTEGER I,J,MAXRBIN,NX(0:127),NS,NE,IX,NTOTAL
C
      XMIN = 0.0
      XMAX = 120.0
      DX = 1.0
C
C     Determine the bin #'s for the limits
      NS = INT(XMIN/DX)
      NE = INT(XMAX/DX)
C
C     Zero things
      DO I=0,127
         NX(I)=0
      ENDDO

C     Place the counts in the arrays
      DO I = 1, MAXRBIN
       DO J = 1, 16
         IX = INT(EAMPS(I,J)/DX)
         IX = MAX(MIN(IX,NE),NS)
         NX(IX) = NX(IX) + 1
         NTOTAL = NTOTAL + 1
       ENDDO
      ENDDO
C
      IX = 0
      J = 0
      AMPV = 0.0
      DO I = NS,NE
         IF (NX(I).GT.J) THEN
            J = NX(I)
            IX = I
         ENDIF
      ENDDO
      AMPV = XMIN + REAL(IX)*DX
      RETURN
      END
C
C====+==================================================================+==
C
      SUBROUTINE SORT_SPECTRA(SORT1,SORT2,SORT3,SORT4,NSRT,  ISEQ)
C
C     This subroutine takes 4 tables of signal characteristics,
C     typically Polarization, Zenith, Azimuth and Doppler,
C     and returns the sequence of channel numbers that sorts these
C     keys in increasing order, from min to max value.
C     NSRT is the number (1-4) of the SORT's to perform
C     This is a simple formatting trick.
C
C     ***WARNING***  This code uses a simple multiplication trick to
C     rank the values relative to each other.  This seems weak and
C     prone to failure.
C
      REAL SORT1(16),SORT2(16),SORT3(16),SORT4(16),SORT(16)
      INTEGER ISEQ(16),NSRT,I
C
      DO I=1,16
         ISEQ(I) = I
      ENDDO
C
      IF ((NSRT.LT.1).OR.(NSRT.GT.4)) RETURN
      DO I=1,16
         SORT(I) = SORT1(I)
         IF(NSRT.GE.2) SORT(I)=1000.*SORT(I)+SORT2(I)
         IF(NSRT.GE.3) SORT(I)=1000.*SORT(I)+SORT3(I)
         IF(NSRT.GE.4) SORT(I)=1000.*SORT(I)+SORT4(I)
      ENDDO
      CALL INDEXX(16,SORT,  ISEQ)

      RETURN
      END      
C
C====+==================================================================+==
C
      SUBROUTINE CHANNEL_MAP(PREFACE, APOL,HDOP,AZ,EL,NBEAM,NDOPP)
C
C     T. Bullett, AFRL, 18 January 2002
C
C     This subroutine performs the rather complex task of supplying the
C     'meta-data' for the 16 channels of the 16 channel Digisonde 256 
C     data.  The meta-data, or data about data, provide information on the
C     context of these different channels.  For the Digisonde 256, the
C     'channel number' can specify Polarization, Doppler and Arrival Angle.
C     It is the precise values of these that this routine computes, based
C     instrument settings.
C     There are lots of caviats here, mostly falling in categoryies:
C       A) That particular setting was not coded or debugged
C       B) Your Digisonde may have different values in its PROMs or
C          was installed differently.
C     The numerical values of polarization are set to +-90.0 to be
C     consistent with the 'chirality' factor of the Dynasonde
C       -90.0  is Ordinary  ; +90.0 is eXtraordinary
C        (J.W. Wright, private communication)
C
C     Output tables are indexed by the channel number(1-16) and contain:
C       APOL - Rx Antenna Polarization Mode O (-90.0 )or X (+90.0)
C       HDOP - Doppler shift, Hz
C       AZ   - Rx antenna beam azimuth, degrees
C       EL   - Rx antenna beam elevation, degrees
C       PZA  - The PZA or 'Meaning' byte defined by UMLCAR (cf Galkin 2001) 
C
C     Internal Variables
C       IBM(4) - Beam Mask to determine which channels are different beams
C       IST(16) - Status Table as a function of channel #
C       IS2RD(4,16) - Status to Relative Doppler Shift
C       IRDOP(16) - Relative Doppler Shift, in 'bins'
C       IB  - Beam #
C       IDD - Doppler Number
C
C     Working with the 4 bit status number, the IDX upper bits
C     are for Doppler, 4-IDX lower bits are IB 'beams'.
C     The relative Doppler index is the upper IDX
C     bits of the Status.  The high bit of these is the sign
C     of the Doppler shift (1=negative).
C     This is embodied in the IS2RD table.
C     Example:  For a T=1 ionogram (O/X and 8 Dopplers), the
C               4 status bits are SDDB where
C               S  = Doppler sign
C               DD = Doppler Magnitude (kind of)
C               B  = Beam number (O or X)
C     
C     * * * W A R N I N G * * * THIS DOES NOT WORK FOR 256 HEIGHTS
C     I HAVE NOT BEEN ABLE TO FIGURE OUT HOW 256 HEIGHT DATA ARE
C     PLACED INTO THE 16 CHANNELS.  I THINK IT IS EVEN/ODD
C
      INTEGER*1 PREFACE(57)
      INTEGER I,IC2S(4,16),NBEAM,NDOPP,IDX,IBM(4),IB,IDD
      INTEGER IST(16),IRDOP(16),IS2RD(4,0:15)
C
      REAL APOL(16),HDOP(16),AZ(16),ZN(16),DFR
      INTEGER PZA(16),S2PZA
      REAL PZA2ZN,PZA2AZ

C
      DATA IBM /7,3,1,0/
C
C----- Channel to Status Table 
      DATA (IC2S(1,I),I=1, 8) / 8, 9,10,11,12,13,14,15/
      DATA (IC2S(1,I),I=9,16) / 0, 1, 2, 3, 4, 5, 6, 7/
      DATA (IC2S(2,I),I=1, 8) / 8, 9,10,11,12,13,14,15/
      DATA (IC2S(2,I),I=9,16) / 4, 5, 6, 7, 0, 1, 2, 3/
      DATA (IC2S(3,I),I=1, 8) / 8, 9,10,11,12,13,14,15/
      DATA (IC2S(3,I),I=9,16) / 6, 7, 4, 5, 2, 3, 0, 1/
      DATA (IC2S(4,I),I=1, 8) / 8, 9,10,11,12,13,14,15/
      DATA (IC2S(4,I),I=9,16) / 7, 6, 5, 4, 3, 2, 1, 0/
C
C     Status to Relative Doppler number.  The first index is 
C     opposite in sense to the previous table.
      DATA (IS2RD(4,I),I=0, 7) /-8, -7, -6, -5, -4, -3, -2, -1/
      DATA (IS2RD(4,I),I=8,15) / 1,  2,  3,  4,  5,  6,  7,  8/
      DATA (IS2RD(3,I),I=0, 7) /-4, -4, -3, -3, -2, -2, -1, -1/ 
      DATA (IS2RD(3,I),I=8,15) / 1,  1,  2,  2,  3,  3,  4,  4/
      DATA (IS2RD(2,I),I=0, 7) /-2, -2, -2, -2, -1, -1, -1, -1/
      DATA (IS2RD(2,I),I=8,15) / 1,  1,  1,  1,  2,  2,  2,  2/
      DATA (IS2RD(1,I),I=0, 7) /-1, -1, -1, -1, -1, -1, -1, -1/
      DATA (IS2RD(1,I),I=8,15) / 1,  1,  1,  1,  1,  1,  1,  1/

      SAVE IC2S,IS2RD
C
C------
C     Get the number and spacing for the doppler lines.
      CALL DOP_D256(PREFACE,  NDOPP,DFR)
      NBEAM = 16/NDOPP
C
C     IDX gives us the Channel-to-Status Table and the 
C     Status-to-Relative-Doppler Table row number to use
C     What I do here is make IDX the log base 2 of NDOPP
      IDX=3
      DO I=1,4
         IF ((NDOPP/2**I).EQ.1) IDX = I
      ENDDO
C
C
      DO ICH=1,16
C        The 'status' value for each channel
         IST(ICH) = IC2S(IDX,ICH)
C        Beam # for this channel
         IB = IAND(IBM(IDX),IST(ICH))
C        Relative Doppler for this 'status'
         IDD = IS2RD(IDX, IST(ICH))
         IRDOP(ICH) = IDD
         HDOP(ICH) = DFR*REAL(IDD)

cdb      write(*,*) 'Ch,St,DD,B',ICH,IST(ICH),IDD,IB,
cdb  +            IRDOP(ICH),HDOP(ICH)
C        Doppler table is complete.
C
C        Now look up the non-Doppler metadata by the Beam#
C        using the 'PZA Byte' concept (c.f. Galkin)
         PZA(ICH) = S2PZA(PREFACE, IST(ICH))
         APOL(ICH) = PZA2POL(PZA(ICH))
C        **** Need to make a PZA2AZ and PZA2ZN
         AZ(ICH) = PZA2AZ(PZA)
         ZN(ICH) = PZA2ZN(PZA)
      ENDDO
C
C
      RETURN
      END

C
C====+==================================================================+==
C
      SUBROUTINE FIXPHASE(RGT,EPHASE,MAXRBIN)
C
C     The Digisondes all perform complex (In-phase and Quadrature)
C     sampling of the receiver output by measuring the 225 kHz IF
C     signal twice at time separation equal to 90 degrees of phase 
C     in the IF.  This is 1.11111 us.
C     A consequence of this is that if successive range gates are not
C     spaced at integer multiples of the IF period (4.444444us) then
C     there will be a phase difference induced by this sampling
C     at the IF.  For dR=1.000 km, this is a 6.6667us dt, or 1.50
C     IF cycles.  For dR=5.000 km, there are 7.500 IF cycles between
C     successive range gates, thus every other range gate is 180 degree
C     phase offset from its neighbors.
C
C     This routine corrects this, referencing all phases to the first
C     range gate and removing the 360 degree miltiples.
C
      INTEGER I,J,MAXRBIN
      REAL R0,DP,EP
      REAL RGT(256),EPHASE(256,16)
C
      R0 = RGT(1)
      DO I=2,MAXRBIN
         DP = 540.0*(RGT(I)-R0)
         DO J=1,16
            EP  = EPHASE(I,J) + DP
            EPHASE(I,J) = EP - 360.0*INT(EP/360.0)
         ENDDO
      ENDDO
      RETURN
      END
C
C
C====+==================================================================+==
C
      REAL FUNCTION PZA2POL(PZA)
C
C     Calculate the polarization based on the PZA byte
C     The numerical values of polarization are set to +-90.0 to be
C     consistent with the 'chirality' factor of the Dynasonde
C        (J.W. Wright, private communication)
C
C     Polarizaion is in the 64's bit of the 0PVZAAAA bits in PZA
      INTEGER PZA
C
      IF (IAND(PZA,64).EQ.0) THEN
C        O Polarization
         PZA2POL = -90.0
      ELSE
C        X polarization
         PZA2POL = +90.0
      ENDIF
      RETURN
      END
C
C====+==================================================================+==
C
      REAL FUNCTION PZA2AZ(PZA)
C
C     Compute the Azimuth Angle from the PZA byte.
C
C     Zenith is in the 16's and 32's bit of the 0PVZAAAA bits in PZA

      INTEGER PZA, IA
C
      PZA2AZ = 0.0
      IA = IAND(PZA,15)
C
      IF (IA.EQ.0) THEN
         PZA2AZ = 0.0
      ELSE IF (IA.LT.13) THEN
         PZA2AZ = 30.0 * REAL(IA-1)
      ELSE
         WRITE(0,*) 'Error in PZA value ', PZA
      ENDIF
      RETURN
      END
C
C
C====+==================================================================+==
C
      REAL FUNCTION PZA2ZN(PZA)
C
C     Compute the Zenith Angle from the PZA byte.
C
C     Zenith is in the 16's and 32's bit of the 0PVZAAAA bits in PZA
C
      INTEGER PZA
C
      PZA2ZN = 0.0
C
      IF (IAND(PZA,32).NE.0) THEN
C        This is overhead
         PZA2ZN = 0.0
      ELSE IF (IAND(PZA,16).EQ.0) THEN
C        Small zenith angles
         PZA2ZN = 11.0
      ELSE IF (IAND(PZA,16).EQ.0) THEN
C        Large zenith angles
         PZA2ZN = 22.0
      ELSE
         WRITE(0,*) 'Error in Zenith value of PZA byte...'
         PZA2ZN = 0.0
      ENDIF
      RETURN
      END
C
C
C====+==================================================================+==
C
      INTEGER FUNCTION S2PZA(PREFACE,ISTAT)
C
C     Convert Status to PZA byte using the ZT tables, Corrected for L
C
C
      INTEGER Z,T,L,ISTAT,TE,ISE,LR
      INTEGER*1 PREFACE(57)

C     Tables in a common block
      INTEGER IZT(0:7,0:7,8)
      COMMON /ZTTABLE/ IZT
C
C     Ignore the high bit on Z
      Z = IAND(PREFACE(46),7)
C
C     Ignore the 4 bit on T
      T = PREFACE(47)
      TE = IAND(T,3) + IAND(T,8)/2

C     The 'effective' status, ignoring the high bit and shifted to
C     the range 1-8
      ISE = IAND(ISTAT,7)+1
C
      S2PZA = IZT(Z,T,ISE)
C
C     Correct for L
      L = PREFACE(45)
      IF ((L.EQ.0).OR.(L.GT.12)) THEN
         WRITE(0,*) 'Warning:  Value of L= ',L,' is unsupported.'
      ELSE
C        The impact of L is to increment the AAAA portion of the 
C        0PVZAAAA byte in S2PZA, with 13 wrapping back to 1, etc.
         LR = IAND(S2PZA,15)
         LR = LR + (L-1)
         IF (LR.GT.12) LR = LR - 12
         S2PZA = IAND(S2PZA,240) + LR
      ENDIF         
      RETURN
      END
C      
C  =======================================================================
C
      SUBROUTINE DOP_D256(PREFACE,   NDOPP,DFR)
C
C     Determines the number and spacing (Doppler Frequency Resolutuion) of
C     the Ionogram Doppler bins.
C     Some principles:
C      - The maximum number of unique Doppler lines is 16 [-8..8]
C      - There is never a zero Doppler in a D256.
C      - The Doppler spectrum is always symmetric about zero

      INTEGER*1 PREFACE(57)
      INTEGER NDOPP,NDT(0:3,4),I,J,IT,IH,IX,IT2,IN,IR,IW
      INTEGER IDC,NB,NPB,ISM
      REAL CIT,DFR
      LOGICAL H8, X123, X04567, X8, IT4
      REAL PRF_D256
C
C     Preface Parameters
      IT = PREFACE(47)
      IT2 = IAND(IT,3)
      IH = PREFACE(54)
      IX = PREFACE(44)
      IN = PREFACE(48)
      IR = PREFACE(49)
      IW = PREFACE(50)
      IT4 = (IAND(IT,4).NE.0)
C
      NDOPP = 0
C
C     This is the enumeration of Table 5.7, p96 in the 'Green Book"
      DO I= 0, 3
         DO J=1,4
            NDT(I,J) = INT(16/2**(I+J))
         ENDDO
      ENDDO
C
C     Some logicals to help decide which column of NDT() to use
      H8 = (IH.GE.8)
      X8 = (IX.GE.8)
      X123 = (IX.EQ.1).OR.(IX.EQ.2).OR.(IX.EQ.3)
      X04567 = (IX.EQ.0).OR.((IX.GE.4).AND.(IX.LE.7))
C
      IF ((.NOT.H8).AND.X04567) THEN
C        H<8;X=0,4,5,6,7
         IDC = 1
      ELSE IF ((.NOT.X8.AND.X123) .OR. (X8.AND.X04567)) THEN
C        H<8;X=1,2,3 -or- H>=8;X=0,4,5,6,7 
         IDC = 2
      ELSE IF (.NOT.(H8.OR.X8).OR.(H8.AND.X123)) THEN
C        X<8;X>=8 or H>=8;X=1,2,3
         IDC = 3
      ELSE IF(H8.AND.X8) THEN
C        H>=8; X>=8
         IDC=4
      ELSE
         IDC = 1
         WRITE(0,*) 'Doppler Table Error..'
      ENDIF
C
      NDOPP = 2*NDT(IT2,IDC)
C
C     NB is the 'Number of Beams' or simultaneous interlaced signals
      IF (NDOPP.EQ.0) THEN
         NB = 0
      ELSE
         NB = 16/NDOPP
      ENDIF
C     NPB is the Number of Pulses per Beam
      NPB = 2**(IN+1)
C     ISM is the Number of Sample adjuster for the W parameter
C                     **** untested ****
      IF (IAND(IW,2).EQ.0) THEN
         ISM = 1
      ELSE
         ISM = 2
      ENDIF

C     Determine the Coherent Integration Time (CIT,sec)
      CIT=REAL(NB*NPB*ISM)/PRF_D256(IR)
C     Now the Doppler Frequency Resolution, which is also the spacing.
      DFR = 1.0
      IF (CIT.NE.0.0) DFR = 0.50/CIT
      IF (IT4) DFR = 2.0*DFR

      RETURN
      END
C
C  =======================================================================
C
      REAL FUNCTION PRF_D256(IR)
C
C     Determine the Pulse Repetition Frequency (PRF) of a D256
C     from the R preface parameter

      INTEGER IR,IR3
C
      IR3 = IAND(IR,3)
C
      IF (IR3.EQ.0) THEN
         PRF_D256 = 50.0
      ELSE IF (IR3.EQ.1) THEN
         WRITE(0,*) "Illegal PRF R= ",IR
         PRF_D256 = 100
      ELSE IF (IR3.EQ.2) THEN
         PRF_D256 = 100.0
      ELSE
         PRF_D256 = 200.0
      ENDIF
      RETURN
      END
C
C  =======================================================================
C
      LOGICAL FUNCTION UNPACK_16C(IBUF,   
     +                    TIME,PREFACE,FREQ,EAMPS,EPHASE,RGT,MAXRBIN)
C
C     This subroutine takes a 4K block of 16 channel amplitude and phase
C     Digisonde 256 data provides the following output.
C        PREFACE - The decoded preface for this ionogram
C        TIME - Time of the observation: YYYY DDD HH:MM:SS
C        FREQ - The precise sounding frequency, in MHz
C        EAMPS(256,16) - Received amplitudes vs range gate and channel # ,
C                        in dB and corrected for all D256 processor settings
C                        but not for any antenna gains, cable losses, etc.
C        EPHASE(256,16)- Received phase vs range and channel #, in degrees
C        RGT(256) - Range Gate Table assigns a range in km to each range gate.
C        MAXRBIN -  Maximum range bin number (128 or 256)
C     Interpetation of the channel number in terms of polarization, Doppler,
C     and Rx antenna beam position (arrival angle) is left to another routine.
C
      REAL FREQ, EAMPS(256,16), EPHASE(256,16), RGT(256)
      CHARACTER*17 TIME
      INTEGER*1 IBUF(4096),PREFACE(57)
      INTEGER MAXRBIN,MAXCHAN, IRTYPE
C
C     Local variables
      INTEGER I,IR,IC,YYYY,ITMP,MASK
      REAL DBSCALE, DEG8
C
C     FUNCTIONS
      REAL FRQ_D256
C
      PARAMETER (DEG8=360.0/256.0,MASK=255)
C
C     Zero the arrays
      DO IR=1,256
         RGT(IR)=0.0
         DO IC=1,16
            EAMPS(IR,IC)=0.0
            EPHASE(IR,IC)=0.0
         ENDDO
      ENDDO
      TIME=""
      FREQ=0.0
      MAXRBIN=0
C
C     Check for 16 channel data type.  Exit if not
      ITMP = IRTYPE(IBUF)
      IF ((ITMP.NE.12).AND.(ITMP.NE.13)) THEN
         UNPACK_16C = .FALSE.
         RETURN
      ENDIF

      CALL PREF_16C(IBUF, PREFACE)
C
      YYYY=10*PREFACE(1)+PREFACE(2)
      IF (YYYY.LT.82) THEN
         YYYY=YYYY+2000
      ELSE
         YYYY=YYYY+1900
      ENDIF
      WRITE(TIME,117) YYYY, (PREFACE(I),I=3,11)
 117  FORMAT (I4.4,1X,3I1,1X,2I1,':',2I1,':',2I1)
C
C
C     Get the frequency value
      FREQ = FRQ_D256(PREFACE)
C
C     Set the range gate values
      CALL RG_D256(PREFACE,  MAXRBIN, RGT)
      MAXCHAN = 16
      IF (MAXRBIN.EQ.256) MAXCHAN=8

C     The amplitude scale on the ionograms, from
C     4 upper bits give 4 or 6 dB per count depending on Z<8
      DBSCALE = 4.0/16.0
      IF (PREFACE(46).LT.8) DBSCALE = 1.5*DBSCALE
C
      AUTOGAIN = 6.0*REAL(PREFACE(53))
C      
C     Extract the amplitudes and phases from the array.
C     They are unsigned 8 bit integers so we have to treat them carefully
C     NOTE:  This won't work for 256 range data
      DO IR=1,MAXRBIN
         DO IC=1,MAXCHAN
            IAX=256*(IC-1)+IR
            ITMP=IAND(MASK,IBUF(IAX))
            EAMPS(IR,IC)=DBSCALE*REAL(ITMP) + AUTOGAIN
            IPX=IAX+128
            ITMP=IAND(MASK,IBUF(IPX))
            EPHASE(IR,IC)=DEG8*REAL(ITMP)
         ENDDO
      ENDDO
C
      UNPACK_16C = .TRUE.
C
C     todo: Move data around for the 256 range gate modes.
C           Need to get some 256 range data for testing!
      RETURN
      END
C
C====+==================================================================+==
C
      REAL FUNCTION FRQ_D256(PREFACE)
C
C     Function retuns the actual radio frequency used for sounding,
C     including the adjustments for frequency search, if enabled.
C     NOTE: This does NOT work for closely spaced frequency mode
C     (aka Precision Group Height mode) 
C
C     ISSUE:  The documentation is inconsistent with itself and with
C             experience.  Actual values observed are 0,1,2 and F
C             For this subroutine F is assumed to be -1
C
      INTEGER*1 PREFACE(57)
      INTEGER I,II,J
      REAL FREQ,DELF
C
      I = PREFACE(56)
C     II is the actual value of I used for this frequency
      II = PREFACE(52)
      IF (II.EQ.15) II = -1
C
C     Frequency digits in in preface locations 20-25
C     PREFACE(20) = 10.0    MHz
C     PREFACE(21) =  1.0    MHz
C     PREFACE(22) =  0.1    MHz (100 kHz)
C     PREFACE(23) =  0.01   MHz ( 10 kHz)
C     PREFACE(24) =  0.001  MHz (  1 kHz)
C     PREFACE(25) =  0.0001 MHz (100 Hz)

      FREQ=0.0
      DO J=1,-4,-1
         FREQ=FREQ + REAL(10.0**J*PREFACE(21-J))
      ENDDO
C
      DELF = 0.00250
      IF (I.EQ.0) THEN
         CONTINUE
      ELSE IF (I.EQ.1) THEN
         DELF = DELF + 0.020
      ELSE IF (I.EQ.2) THEN
         DELF = DELF + 0.010*REAL(II)
      ELSE IF (I.EQ.3) THEN
         DELF = DELF + 0.020*REAL(II)
      ENDIF
C
      FRQ_D256 = FREQ + DELF
cdb      write(*,'(A,F8.5,1X,I2,1X,I2)') 'F,I,I*= ', FRQ_D256,I,II
      RETURN
      END
C
C====+==================================================================+==
C
      SUBROUTINE RG_D256(PREFACE,  NRG, RGT)
C
C     Given the PREFACE, returns the number of range gates
C     and their values
C
      INTEGER*1 PREFACE(57)
      INTEGER NRG
      REAL RGT(256)
C
C
      INTEGER IH, IE, IBL
      REAL HTBL(0:15),ETBL(5)
C
C
      DATA HTBL /2.5, 5.0, 10.0, 2.5, 0.0, 0., 0., 0.,
     +           2.5, 5.0, 10.0, 2.5, 5.0, 0., 0., 0. /
      DATA ETBL /10.0, 60.0, 160.0, 380.0, 760.0/
C
C     The range gates are determined by preface parameters H and E
C     as defined in the Digisonde 256 "Green Book"
      IH = PREFACE(54)
      IE = PREFACE(55)
C
      IF ((IH.LT.0).OR.(IH.GT.15).OR.(IE.LT.1).OR.IE.GT.5) THEN
         WRITE(0,*) "UNDEFIEND RANGE GATES H=",IH," E=",IE
         NRG = 0
         RETURN
      ENDIF
C
      IF (IH.GE.8) THEN
         NRG = 256
      ELSE
         NRG = 128
      ENDIF
C
C     The standard linear range gate modes
      DO I=1,NRG
         RGT(I) = ETBL(IE) + REAL(I-1)*HTBL(IH)
      ENDDO
C
C     Now the bi-linear modes. IBL is the # of gates at the lower ranges
      IBL = 0
      IF (IH.EQ.3) THEN
         IBL = 40
      ELSE IF ((IH.EQ.11).OR.(IH.EQ.12)) THEN
         IBL = 128
      ENDIF
C
C     For all these modes the RG spacing is double for the upper gates
      IF (IBL.EQ.0) RETURN
      DO I = 1, IBL
         RGT(I) = ETBL(IE) + REAL(I-1)*HTBL(IH)
      ENDDO
      DO I = IBL+1, NRG
         RGT(I) = ETBL(IE) + 2.0*REAL(I-1)*HTBL(IH)
      ENDDO
C
      RETURN
      END
C
C  ========================================================================
C
      SUBROUTINE ZTL_INIT()
C
C     Initialize the ZT table
C     This routine sets the values for the ZT table used for
C     determining Rx antenna beams, etc.  It is initialized this way
C     to allow for a custom table to be loaded early in the main program,
C     perhaps from a file.
C
C     The Meaning byte is defined as 0PVZAAAA where
C        P = Polarization (0=Ordinary, 1=eXtraordinary)
C        V = Vertical  (0=Not vertical, 1 = Vertical)
C        Z = Zenith Angles (0 for small(11deg), 1 for Large (22deg))
C        AAAA = Azimuth Value (0-12)
C
C     ToDo: Allow this to optionally read from an UMLCAR format ZT file.
C
C     Source: Green Book table 5.9, P.98, Revised 11Jun86,5Nov87,28Dec88
C

      INTEGER IZT(0:7,0:7,8)
      COMMON /ZTTABLE/ IZT
C
C    Z=0,8 T=0,1,2,3, 8,9,A,B
      DATA (IZT(0,0,I),I=1,8) /32, 32, 32, 32, 32, 32, 32, 32/
      DATA (IZT(0,1,I),I=1,8) /17, 17, 17, 17, 17, 17, 17, 17/
      DATA (IZT(0,2,I),I=1,8) /17, 18, 96, 32, 17, 18, 96, 32/
      DATA (IZT(0,3,I),I=1,8) /17, 18, 19, 20, 21, 22, 96, 32/
      DATA (IZT(0,4,I),I=1,8) /17, 17, 17, 17, 17, 17, 17, 17/
      DATA (IZT(0,5,I),I=1,8) /17, 18, 17, 18, 17, 18, 17, 18/
      DATA (IZT(0,6,I),I=1,8) /17, 18, 19, 20, 17, 18, 19, 20/
      DATA (IZT(0,7,I),I=1,8) /17, 18, 19, 20, 21, 22, 23, 24/
C    Z=1,9 T=0,1,2,3, 8,9,A,B
      DATA (IZT(1,0,I),I=1,8) / 0,  0,  0,  0,  0,  0,  0,  0/
      DATA (IZT(1,1,I),I=1,8) /17,  1, 17,  1, 17,  1, 17,  1/
      DATA (IZT(1,2,I),I=1,8) /17, 17, 32, 32, 17, 17, 32, 32/
      DATA (IZT(1,3,I),I=1,8) /17, 19, 21, 23, 25, 27, 96, 32/
      DATA (IZT(1,4,I),I=1,8) / 0,  0,  0,  0,  0,  0,  0,  0/
      DATA (IZT(1,5,I),I=1,8) /17, 19, 17, 19, 17, 19, 17, 19/
      DATA (IZT(1,6,I),I=1,8) /17, 17, 17, 17, 17, 17, 17, 17/
      DATA (IZT(1,7,I),I=1,8) /17, 17, 17, 17, 19, 19, 19, 19/
C    Z=2,A T=0,1,2,3, 8,9,A,B
      DATA (IZT(2,0,I),I=1,8) / 0,  0,  0,  0,  0,  0,  0,  0/
      DATA (IZT(2,1,I),I=1,8) /17, 32, 17, 32, 17, 32, 17, 32/
      DATA (IZT(2,2,I),I=1,8) /81, 81, 17, 17, 81, 81, 17, 17/
      DATA (IZT(2,3,I),I=1,8) /81, 81, 81, 81, 17, 17, 17, 17/
      DATA (IZT(2,4,I),I=1,8) / 0,  0,  0,  0,  0,  0,  0,  0/
      DATA (IZT(2,5,I),I=1,8) /17, 20, 17, 20, 17, 20, 17, 20/
      DATA (IZT(2,6,I),I=1,8) /17, 20, 23, 26, 17, 20, 23, 26/
      DATA (IZT(2,7,I),I=1,8) /17, 17, 17, 17, 32, 32, 32, 32/
C    Z=3,B T=0,1,2,3, 8,9,A,B
      DATA (IZT(3,0,I),I=1,8) /81, 81, 81, 81, 81, 81, 81, 81/
      DATA (IZT(3,1,I),I=1,8) /81, 17, 81, 17, 81, 17, 81, 17/
      DATA (IZT(3,2,I),I=1,8) /17, 23, 96, 32, 17, 23, 96, 32/
      DATA (IZT(3,3,I),I=1,8) /17, 17, 23, 23, 96, 96, 32, 32/
      DATA (IZT(3,4,I),I=1,8) / 0,  0,  0,  0,  0,  0,  0,  0/
      DATA (IZT(3,5,I),I=1,8) /17, 23, 17, 23, 17, 23, 17, 23/
      DATA (IZT(3,6,I),I=1,8) /17, 23,  1,  7, 17, 21,  1,  7/
      DATA (IZT(3,7,I),I=1,8) /17, 17, 17, 17, 23, 23, 23, 23/
C    Z=4,C T=0,1,2,3, 8,9,A,B
      DATA (IZT(4,0,I),I=1,8) /96, 96, 96, 96, 96, 96, 96, 96/
      DATA (IZT(4,1,I),I=1,8) / 1,  1,  1,  1,  1,  1,  1,  1/
      DATA (IZT(4,2,I),I=1,8) / 1,  2, 96, 32,  1,  2, 96, 32/
      DATA (IZT(4,3,I),I=1,8) / 1,  2,  3,  4,  5,  6, 96, 32/
      DATA (IZT(4,4,I),I=1,8) / 1,  1,  1,  1,  1,  1,  1,  1/
      DATA (IZT(4,5,I),I=1,8) /17, 19, 17, 19, 17, 19, 17, 19/
      DATA (IZT(4,6,I),I=1,8) /17, 17, 17, 17, 17, 17, 17, 17/
      DATA (IZT(4,7,I),I=1,8) /17, 17, 17, 17, 19, 19, 19, 19/
C    Z=5,D T=0,1,2,3, 8,9,A,B
      DATA (IZT(5,0,I),I=1,8) / 0,  0,  0,  0,  0,  0,  0,  0/
      DATA (IZT(5,1,I),I=1,8) /96, 32, 96, 32, 96, 32, 96, 32/
      DATA (IZT(5,2,I),I=1,8) /96, 96, 32, 32, 96, 96, 32, 32/
      DATA (IZT(5,3,I),I=1,8) / 1,  3,  5,  7,  9, 11, 96, 32/
      DATA (IZT(5,4,I),I=1,8) / 0,  0,  0,  0,  0,  0,  0,  0/
      DATA (IZT(5,5,I),I=1,8) / 1,  3,  1,  3,  1,  3,  1,  3/
      DATA (IZT(5,6,I),I=1,8) / 1,  1,  1,  1,  1,  1,  1,  1/
      DATA (IZT(5,7,I),I=1,8) / 1,  1,  1,  1,  3,  3,  3,  3/
C    Z=6,E T=0,1,2,3, 8,9,A,B
      DATA (IZT(6,0,I),I=1,8) / 0,  0,  0,  0,  0,  0,  0,  0/
      DATA (IZT(6,1,I),I=1,8) / 1, 32,  1, 32,  1, 32,  1, 32/
      DATA (IZT(6,2,I),I=1,8) /65, 65,  1,  1, 65, 65,  1,  1/
      DATA (IZT(6,3,I),I=1,8) /65, 65, 65, 65,  1,  1,  1,  1/
      DATA (IZT(6,4,I),I=1,8) / 0,  0,  0,  0,  0,  0,  0,  0/
      DATA (IZT(6,5,I),I=1,8) / 1,  4,  1,  4,  1,  4,  1,  4/
      DATA (IZT(6,6,I),I=1,8) / 1,  4,  7, 10,  1,  4,  7, 10/
      DATA (IZT(6,7,I),I=1,8) / 1,  1,  1,  1, 17, 17, 17, 17/
C    Z=7,F T=0,1,2,3, 8,9,A,B
      DATA (IZT(7,0,I),I=1,8) /65, 65, 65, 65, 65, 65, 65, 65/
      DATA (IZT(7,1,I),I=1,8) /65,  1, 65,  1, 65,  1, 65,  1/
      DATA (IZT(7,2,I),I=1,8) / 1,  7, 96, 32,  1,  7, 96, 32/
      DATA (IZT(7,3,I),I=1,8) /96, 96, 96, 96, 32, 32, 32, 32/
      DATA (IZT(7,4,I),I=1,8) / 0,  0,  0,  0,  0,  0,  0,  0/
      DATA (IZT(7,5,I),I=1,8) / 1,  7,  1,  7,  1,  7,  1,  7/
      DATA (IZT(7,6,I),I=1,8) /32, 32, 32, 32, 32, 32, 32, 32/
      DATA (IZT(7,7,I),I=1,8) / 1,  1,  1,  1,  7,  7,  7,  7/
C
C     Insert code here to read from sdtin or a file

      RETURN
      END
C
C  ========================================================================
C
      LOGICAL FUNCTION READ4KB(IU,  IBUF)
C
C     Subroutine to read 4096 bytes from opened file unit IU
C     Returns the values in IBUF
C     Return value is TRUE if there is some problem, such as EOF
C     Specialized for Linux g77 compiler which does not easily support
C       binary files
C
      INTEGER I,IER,IB
      INTEGER*1 IBUF(4096)
      DATA IB /0/
      SAVE IB
C
      IB = IB + 1
      DO I=1,4096
         IBUF(I)=0
      ENDDO
      READ(IU,IOSTAT=IER,REC=IB) (IBUF(I),I=1,4096)
      READ4KB = (IER.NE.0)

cdb      write(*,101) IB, (IBUF(I),i=1,20)
cdb  101  format ('Read-> ',I4,'->',20(1X,Z2))
      RETURN
      END
C
C   ======================================================================
C
	INTEGER FUNCTION IRTYPE(IBUF)
C
C	This function takes the Digisonde 256/DPS data block in IBUF and
C	determines the actual data type (Ionogram, Ionogram continuation,
C	Drift, Artist, Raw Ionogram) and returns it as the function value.
C	If the data block is not one of these types (ie unknown) the 
C	function returns a zero value for the recordtype.
C
C       NOTE: Due to a bug in the recording of 16 channel data in Artist3
C         circa 2001, where the record type is not encoded into the 4 lowest
C         bits of the first byte of every 8K block, this routine must be
C         specific to 16 channel data and also VERY FRAGILE!
C
C       Empirically, this function returns a recordtype of 1 for what is
C       supposed to be Recordtype 0x0C.
C       Unknown what happens for Recordtype 0x0D
C
	INTEGER IRT,I,KRT
	INTEGER*1 M4L,M1L,ICC,IBUF(4096)
	LOGICAL OK
	PARAMETER (M4L=15,M1L=1,ICC=204)
C
	OK = .FALSE.
C 17Jan2002  TWB -  No record type in data at all!
C Fugly Hacked around this bug, assume  recordtype is 12
        ok = .true.
        IRTYPE = 12
        RETURN
C
C	Make the first cut at the record type.
	IRT = IAND(IBUF(1),M4L)
C	Eliminate false alarms.
C
        KRT=0
	IF ((IRT.EQ.9).OR.(IRT.EQ.8)) THEN
C	   Verify that this is ionogram data.
C	   The third character in the block should be zero.
	   OK = (IBUF(3).EQ.0)
	ELSE IF ((IRT.EQ.10).OR.(IRT.EQ.12).OR.
     +           (IRT.EQ.13)) THEN
C	   For raw data types, extract the record type from the least
C	   significant bits of the first 4 words, and compare that with
C	   the record type. This is less absolute than the Ionogram method.
	   KRT = 0
	   DO I = 0, 3
	      KRT = IOR(KRT,ISHFT(IAND(IBUF(I+1),M1L),I))
	   ENDDO
	   OK = (IRT.EQ.KRT)
	ELSE IF (IRT.EQ.15) THEN
C	   For ARTIST, the 4th and 5th bytes should be Hex CC
	   OK = (IBUF(4).EQ.ICC).AND.(IBUF(5).EQ.ICC)
	ENDIF
C
cdb        write(*,*) 'irtype -> ',IRT,KRT
	IRTYPE = 0
	IF (OK) IRTYPE = IRT
C
	RETURN
	END
C
C =============================================================================
C
      SUBROUTINE PREF_16C(IBUF,  IPREF)
C
C     Extract the preface encoded into the LSB of the amplitude data in the
C     16 channel data blocks.
C     Preface characters 1-31 found in bytes    5-128 for both data types
C     Preface characters 32-57 found in bytes 257-360 for IRTYPE=12
C     Preface characters 32-64 found in bytes 129-232 for IRTYPE=???


      INTEGER I,IP,IE,IRT,IPEND
      INTEGER*1 IBUF(4096),IPREF(57),M1L
      INTEGER IRTYPE
      PARAMETER(M1L=1)

      IRT=IRTYPE(IBUF)
      IPEND=57
      DO IP = 1,IPEND
         IPREF(IP) = 0
         IE = 4*IP+1
c         IF ((IP.GT.31).AND.(IRT.EQ.12)) IE = IE + 128
         IF ((IP.GT.31)) IE = IE + 128
         DO I = 0,3
            IPREF(IP)=IOR(IPREF(IP),ISHFT(IAND(IBUF(IE+I),M1L),I))
         ENDDO
      ENDDO
      RETURN
      END
C
C =============================================================================
C
C     Sort the array ARRIN, Length N, returning the order of the elements
C     indexed in array INDX
C
      SUBROUTINE INDEXX(N,ARRIN,INDX)
      DIMENSION ARRIN(N),INDX(N)
C
      DO  J=1,N
        INDX(J)=J
      ENDDO
      L=N/2+1
      IR=N
      DO WHILE (.TRUE.)
        IF(L.GT.1)THEN
          L=L-1
          INDXT=INDX(L)
          Q=ARRIN(INDXT)
        ELSE
          INDXT=INDX(IR)
          Q=ARRIN(INDXT)
          INDX(IR)=INDX(1)
          IR=IR-1
          IF(IR.EQ.1)THEN
            INDX(1)=INDXT
            RETURN
          ENDIF
        ENDIF
        I=L
        J=L+L
        DO WHILE (J.LE.IR)
          IF(J.LT.IR)THEN
            IF(ARRIN(INDX(J)).LT.ARRIN(INDX(J+1)))J=J+1
          ENDIF
          IF(Q.LT.ARRIN(INDX(J)))THEN
            INDX(I)=INDX(J)
            I=J
            J=J+J
          ELSE
            J=IR+1
          ENDIF
        ENDDO
        INDX(I)=INDXT
      ENDDO

      END










































